from ..torchio import INTENSITY
from ..utils import is_image_dict
from .transform import Transform


class ZNormalization(Transform):
    """
    Subtract mean and divide by standard deviation
    """
    def apply_transform(self, sample):
        for image_dict in sample.values():
            if not is_image_dict(image_dict):
                continue
            if image_dict['type'] != INTENSITY:
                continue
            image_dict['data'] = znorm(image_dict['data'])
        return sample


def znorm(data, masking_function=None):
    if masking_function is None:
        masking_function = mean_plus
    mask_data = masking_function(data)
    values = data[mask_data]
    mean, std = values.mean(), values.std()
    data = data - mean
    data = data / std
    return data


def mean_plus(data):
    return data > data.mean()



import torch
import numpy as np
from ..utils import is_image_dict
from .random_transform import RandomTransform


class RandomFlip(RandomTransform):
    def __init__(
            self,
            axes=(0,),
            flip_probability=0.5,
            seed=None,
            verbose=False,
            ):
        super().__init__(seed=seed, verbose=verbose)
        self.axes = axes
        assert flip_probability > 0
        assert flip_probability <= 1
        self.flip_probability = flip_probability

    def apply_transform(self, sample):
        axes_to_flip_hot = self.get_params(self.axes, self.flip_probability)
        sample['random_flip'] = axes_to_flip_hot
        for image_dict in sample.values():
            if not is_image_dict(image_dict):
                continue
            tensor = image_dict['data']
            dims = []
            for dim, flip_this in enumerate(axes_to_flip_hot):
                if not flip_this:
                    continue
                actual_dim = dim + 1  # images are 4D
                dims.append(actual_dim)
            tensor = torch.flip(tensor, dims=dims)
            image_dict['data'] = tensor
        return sample

    @staticmethod
    def get_params(axes, probability):
        axes_hot = [False, False, False]
        for axis in axes:
            random_number = torch.rand(1)
            flip_this = bool(probability > random_number)
            axes_hot[axis] = flip_this
        return axes_hot



import torch
import SimpleITK as sitk
from ..torchio import LABEL
from ..utils import is_image_dict
from .interpolation import Interpolation
from .random_transform import RandomTransform


class RandomElasticDeformation(RandomTransform):
    def __init__(
            self,
            num_control_points=4,
            deformation_std=15,
            proportion_to_augment=0.5,
            spatial_rank=3,
            image_interpolation=Interpolation.LINEAR,
            seed=None,
            verbose=False,
            ):
        super().__init__(seed=seed, verbose=verbose)
        self._bspline_transformation = None
        self.num_control_points = max(num_control_points, 2)
        self.deformation_std = max(deformation_std, 1)
        self.proportion_to_augment = proportion_to_augment
        self.spatial_rank = spatial_rank
        self.image_interpolation = image_interpolation

    def apply_transform(self, sample):
        bspline_params = None
        sample['random_elastic_deformation'] = {}
        params_dict = sample['random_elastic_deformation']
        for image_dict in sample.values():
            if not is_image_dict(image_dict):
                continue
            if image_dict['type'] == LABEL:
                interpolation = Interpolation.NEAREST
            else:
                interpolation = self.image_interpolation
            # TODO: assert that all images have the same shape
            if bspline_params is None:
                image = self.nib_to_sitk(
                    image_dict['data'][0],
                    image_dict['affine'],
                )
                do_augmentation, bspline_params = self.get_params(
                    image,
                    self.num_control_points,
                    self.deformation_std,
                    self.proportion_to_augment,
                )
                params_dict['bspline_params'] = bspline_params
                params_dict['do_augmentation'] = int(do_augmentation)
                if not do_augmentation:
                    return sample
            image_dict['data'] = self.apply_bspline_transform(
                image_dict['data'],
                image_dict['affine'],
                bspline_params,
                interpolation,
            )

        return sample

    @staticmethod
    def get_params(image, num_control_points, deformation_std, probability):
        mesh_shape = 3 * (num_control_points,)
        bspline_transform = sitk.BSplineTransformInitializer(image, mesh_shape)
        default_params = bspline_transform.GetParameters()
        bspline_params = torch.rand(len(default_params)) * deformation_std
        do_augmentation = torch.rand(1) < probability
        return do_augmentation, bspline_params.numpy()

    @staticmethod
    def get_bspline_transform(image, num_control_points, bspline_params):
        mesh_shape = 3 * (num_control_points,)
        bspline_transform = sitk.BSplineTransformInitializer(image, mesh_shape)
        bspline_transform.SetParameters(bspline_params.tolist())
        return bspline_transform

    def apply_bspline_transform(
            self,
            array,
            affine,
            bspline_params,
            interpolation: Interpolation,
            ):
        if array.ndim != 4:
            message = (
                'Only 4D images (channels, i, j, k) are supported,'
                f' not {array.shape}'
            )
            raise NotImplementedError(message)
        for i, channel_array in enumerate(array):  # use sitk.VectorImage?
            image = self.nib_to_sitk(channel_array, affine)
            bspline_transform = self.get_bspline_transform(
                image,
                self.num_control_points,
                bspline_params,
            )

            resampler = sitk.ResampleImageFilter()
            resampler.SetInterpolator(interpolation.value)
            resampler.SetReferenceImage(image)
            resampler.SetDefaultPixelValue(0)  # should I change this?
            resampler.SetTransform(bspline_transform)
            resampled = resampler.Execute(image)

            channel_array = sitk.GetArrayFromImage(resampled)
            channel_array = channel_array.transpose()  # ITK to NumPy
            array[i] = torch.from_numpy(channel_array)
        return array



from pathlib import Path
from collections.abc import Sequence
from torch.utils.data import Dataset
from ..utils import get_stem
from ..io import read_image, write_image


class ImagesDataset(Dataset):
    def __init__(
            self,
            subjects_list,
            transform=None,
            verbose=False,
            ):
        """
        Each element of subjects_list is a dictionary:
        subject = {
            'one_image': dict(
                path=path_to_one_image,
                type=torchio.INTENSITY,
            ),
            'another_image': dict(
                path=path_to_another_image,
                type=torchio.INTENSITY,
            ),
            'a_label': dict(
                path=path_to_a_label,
                type=torchio.LABEL,
            ),
        }
        See examples/example_multimodal.py for -obviously- an example.
        """
        self.parse_subjects_list(subjects_list)
        self.subjects_list = subjects_list
        self.transform = transform
        self.verbose = verbose

    def __len__(self):
        return len(self.subjects_list)

    def __getitem__(self, index):
        subject_dict = self.subjects_list[index]
        sample = {}
        for image_name, image_dict in subject_dict.items():
            image_path = image_dict['path']
            tensor, affine = self.load_image(image_path)
            image_sample_dict = dict(
                data=tensor,
                path=str(image_path),
                affine=affine,
                stem=get_stem(image_path),
                type=image_dict['type'],
            )
            sample[image_name] = image_sample_dict

        # Apply transform (this is usually the bottleneck)
        if self.transform is not None:
            sample = self.transform(sample)
        return sample

    def load_image(self, path):
        if self.verbose:
            print(f'Loading {path}...')
        tensor, affine = read_image(path)
        if self.verbose:
            print(f'Loaded array with shape {tensor.shape}')
        return tensor, affine

    @staticmethod
    def parse_subjects_list(subjects_list):
        def parse_path(path):
            try:
                path = Path(path).expanduser()
            except TypeError:
                print(f'Conversion to path not possible for variable: {path}')
                raise
            return path.is_file()
        if not isinstance(subjects_list, Sequence):
            raise TypeError(
                f'Subject list must be a sequence, not {type(subjects_list)}')
        if not subjects_list:
            raise ValueError('Subjects list is empty')
        for element in subjects_list:
            if not isinstance(element, dict):
                raise TypeError(
                    f'All elements must be dictionaries, not {type(element)}')
            if not element:
                raise ValueError(f'Element seems empty: {element}')
            subject_dict = element
            for image_name, image_dict in subject_dict.items():
                if not isinstance(image_dict, dict):
                    raise TypeError(
                        f'Type {type(image_dict)} found for {image_name},'
                        ' instead of type dict'
                    )
                for key in ('path', 'type'):
                    if key not in image_dict:
                        raise KeyError(
                            f'"{key}" key not found'
                            f' in image dict {image_dict}')
                path = image_dict['path']
                if not parse_path(path):
                    message = (
                        f'File for image "{image_name}"'
                        f' not found: "{path}"'
                    )
                    raise FileNotFoundError(message)

    @staticmethod
    def save_sample(sample, output_paths_dict):
        for key, output_path in output_paths_dict.items():
            tensor = sample[key]['data'][0]  # remove channels dim
            affine = sample[key]['affine']
            write_image(tensor, affine, output_path)
